"""
Engines perform analyses on the tree. They are all run concurrently for performance.

quick wins
- [DONE] detect redundant dirs [S2.a]
- [DONE] detect obvious folders e.g tiff/*.tif* [S2.b]
- [DONE] detect excessive files per directory [S2.c]
- [DONE] detect directories with mixed files [S3.a]
- detect cryptic names (against a dictionary) [N1]
- [DONE] detect dates in names [N1.b]
- [DONE] detect accessions e.g. 'EMPIAR' [N1.c]
- [DONE] detect mixed case in names [N2.a]
- [DONE] detect periods in names [N2.a]
- [DONE] detect odd characters in names [N2.b]
- [DONE] detect long names [N2.d]
- detect inconsistent names [N3.a]
- [DONE] detect external references e.g. 'figure' [N3.c]
- [DONE] detect words to avoid e.g. 'files', 'data' [N3.c]
- detect missing padding [N3.e]
- detect embedded paths (needs file format library) [N3.f]
- [DONE] detect unknown extensions [M1.a]
- [DONE] detect proprietary extensions [M1.b]
- detect presence of documentation e.g. README [M2.a]
- detect presence of checksums [M3]
- detect hard links
- detect symbolic links
- detect broken symbolic links
"""

import shutil

import styled

import bandbox

width, height = shutil.get_terminal_size((80, 60))
RIGHT_COL_WIDTH = 40
LEFT_COL_WIDTH = width - RIGHT_COL_WIDTH - 1


async def _report(dirs: list, rule_text: str, fail_text: str = '', args=None) -> None:
    """Reporting function"""
    summarise = getattr(args, 'summarise', False)
    summarise_size = getattr(args, 'summarise_size', 5)
    print(styled.Styled(f"[[ '{rule_text.ljust(LEFT_COL_WIDTH)}'|bold ]]"), end=" ")
    if dirs:
        if fail_text:
            print(styled.Styled(f"[[ '{fail_text.rjust(RIGHT_COL_WIDTH)}'|fg-red:bold ]]"))
        else:
            fail_text = f"[{len(dirs)} directories] nok".rjust(RIGHT_COL_WIDTH)
            print(styled.Styled(f"[[ '{fail_text}'|fg-red:bold ]]"))
        if summarise and len(dirs) > summarise_size:
            dirs_ = dirs[:summarise_size]
        else:
            dirs_ = dirs
        for item in dirs_:
            print(f"  * {item}")
        if summarise and len(dirs) > summarise_size:
            print(f"  * [+{len(dirs) - summarise_size} others (remove --summarise option to view the full list)]")

    else:
        ok_text = "ok".rjust(RIGHT_COL_WIDTH)
        print(styled.Styled(f"[[ '{ok_text}'|fg-green:bold ]]"))
    print()


async def s2_detect_redundant_directories(tree, args):
    """Detect the presence of redundant directories

    A directory is redundant if:
    - it is empty
    - it is non-empty but is the only child of a parent folder
    """
    # empty folders
    empty_folders = tree.find_empty_directories(include_root=args.include_root)
    await _report(empty_folders, f"{'structure warning':<17} => - redundant directories...", args=args)


async def s2_detect_obvious_directories(tree, args):
    """Detect obvious folders"""
    obvious_folders = tree.find_obvious_directories(include_root=args.include_root)
    await _report(obvious_folders, f"{'structure warning':<17} => - obvious directory names...", args=args)


async def s2_detect_excessive_files_per_directory(tree, args):
    """Detect excessive files per directory"""
    excess_files = tree.find_excessive_files_per_directory()
    await _report(excess_files, f"{'structure warning':<17} => - excessives (>{args._configs.getint('bandbox', 'max_files')}) files per directory...", args=args)


async def s3_detect_directories_with_mixed_files(tree, args):
    """Detect folders with mixed files"""
    mixed_files = tree.find_directories_with_mixed_files()
    await _report(mixed_files, f"{'structure warning':<17} => - directories with mixed files...", args=args)


async def n2_detect_long_names(tree, args):
    """Detect entities with very long names"""
    # print(f"info: working on {tree} with {args}...")
    dirs = tree.find_long_names()
    await _report(dirs, f"{'name warning':<17} => - long names (>{args._configs.getint('bandbox', 'max_name_length')} chars)...", args=args)


async def n1_detect_dates_in_names(tree, args):
    """Detect dates of various formats in names"""
    dirs = tree.find_with_date_names()
    await _report(dirs, f"{'name warning':<17} => - entities with dates in names...", args=args)


async def n1_detect_accessions_in_names(tree, args):
    """Detect accessions in names"""
    dirs = tree.find_accessions_in_names()
    await _report(dirs, f"{'name warning':<17} => - accessions in names...", args=args)


async def n2_detect_mixed_case(tree, args):
    """Detect mixed case"""
    dirs = tree.find_mixed_case()
    await _report(dirs, f"{'name warning':<17} => - mixed case in names...", args=args)


async def n2_detect_odd_characets_in_names(tree, args):
    dirs = tree.find_odd_characters_in_names()
    await _report(dirs, f"{'name warning':<17} => - odd characters [one of '{args._configs.get('bandbox', 'odd_chars')}'] in names...", args=args)


async def n2_detect_excessive_periods_in_names(tree, args):
    dirs = tree.find_excessive_periods_in_names()
    await _report(dirs, f"{'name warning':<17} => - excessive periods in names...", args=args)


async def n3_detect_external_references_in_names(tree, args):
    dirs = tree.find_external_references_in_names()
    await _report(dirs, f"{'name warning':<17} => - external references in names...", args=args)


async def m1_detect_unknown_file_extensions(tree, args):
    dirs = tree.find_unknown_file_extensions()
    await _report(dirs, f"{'misc. warning':<17} => - unknown file extensions...", args=args)


async def n2_detect_non_ascii_characters_in_names(tree, args):
    dirs = tree.find_non_ascii_characters()
    await _report(dirs, f"{'name warning':<17} => - non-ascii characters in names...", args=args)



import os
import pathlib
import typing
import sys


def scandir_recursive(path: pathlib.Path, recursive=True) -> typing.Generator:
    """Recursively scan a directory

    Recursion can be switchef off
    """
    with os.scandir(path) as dir_entries:
        for dir_entry in dir_entries:
            if dir_entry.is_dir():
                if os.listdir(dir_entry.path):
                    yield dir_entry
                    if recursive:
                        yield from scandir_recursive(dir_entry.path)
                else:
                    yield dir_entry
            else:
                yield dir_entry



"""
bandbox analyse # gives suggestions of possible violations
bandbox view    # render tree to show files and folders


inputs:
- file with tree
- path
- filtered path
output:
- display tree
- analyses
- save tree
"""
import configparser
import io
import os
import pathlib
import sys
import types
import unittest

import requests

from bandbox import cli, models, utils, managers

BASE_DIR = pathlib.Path("/Users/pkorir/PycharmProjects/bandbox")
TEST_DATA = BASE_DIR / "test_data"
TEST_CONFIG = BASE_DIR / "bandbox.cfg"


class Tests(unittest.TestCase):
    """Base test class"""

    def setUp(self) -> None:
        # envvars
        os.environ['BANDBOX_CONFIG'] = str(TEST_CONFIG)


class Test(Tests):
    def test_gist(self):
        """Get JSON data from Github gist"""
        response = requests.get(
            "https://gist.githubusercontent.com/paulkorir/5b71f57f7a29391f130e53c24a2db3fb/raw/bandbox.json")
        self.assertEqual(200, response.status_code)
        self.assertIsNotNone(response.text)
        self.assertIsInstance(response.json(), dict)
        self.assertTrue("file_formats" in response.json())


class TestCLI(Tests):
    """BDD for the CLI"""

    def test_view(self):
        """View the tree"""
        args = cli.cli(f"bandbox view")
        self.assertEqual('view', args.command)
        self.assertEqual(pathlib.Path('.'), args.path)
        self.assertEqual('', args.prefix)
        self.assertFalse(args.verbose)
        self.assertTrue(args.hide_file_counts)
        self.assertIsNone(args.input_file)
        self.assertIsNotNone(args.config_file)

    def test_analyse(self):
        """Analyse the given path"""
        args = cli.cli(f"bandbox analyse")
        self.assertEqual('analyse', args.command)
        self.assertEqual(pathlib.Path('.'), args.path)
        self.assertFalse(args.include_root)
        self.assertFalse(args.summarise)
        self.assertEqual(5, args.summarise_size)
        self.assertFalse(args.verbose)
        self.assertIsNotNone(args.config_file)


class TestCore(Tests):
    """Test core entities"""

    def test_tree_init(self):
        """Initialise a tree"""
        tree = models.Tree()
        self.assertIsInstance(tree, models.Tree)
        self.assertTrue(hasattr(tree, 'data'))  # as a UserDict subclass this is where the data sits
        self.assertTrue(hasattr(tree, 'sep'))
        self.assertTrue(hasattr(tree, 'show_file_counts'))

    def test_tree_insert(self):
        """Test inserting paths into the tree"""
        args = cli.cli(f"bandbox view")
        expected_output = dict(
            empty_folder={'empty_folder': {'folder': {}}},
            folder_with_multiple_file_types={
                'folder_with_multiple_file_types': {
                    'folder': {
                        'files': [f"file{i}.tif" for i in range(1, 11)] + [f"file{j}.txt" for j in range(1, 11)]}}
            },
            folder_with_multiple_files={
                'folder_with_multiple_files': {'folder': {'files': [f"file{j}.txt" for j in range(1, 11)]}}
            },
            folder_with_multiple_folders={
                'folder_with_multiple_folders': {
                    f"folder{i}": [f"file{j}.tif" for j in range(1, 11)] for i in range(1, 6)
                }
            },
            folder_with_single_file={'folder_with_single_file': {'folder': {'files': ['file.txt']}}},
            single_empty_folder={'single_empty_folder': {'folder': {}}},
        )
        for data_name, data in expected_output.items():
            dir_entries = utils.scandir_recursive(TEST_DATA / data_name)
            tree = models.Tree.from_data(dir_entries, prefix=str(TEST_DATA), args=args)
            # the best we can do is compare keys
            self.assertListEqual(list(data.keys()), list(tree.data.keys()))

    def test_all_find_methods(self):
        """Test all find methods

        There is no need to write a method for each when the only thing that changes is the data. Rather, we compile
        all the data into a dictionary and only update the dictionary as we add new find methods.

        Each find method on the tree will need:
        - the method name
        - the name of the source folder
        - the expected output
        - a tree instance (created on the fly)
        """
        args = cli.cli(f"bandbox analyse")
        data = [
            {
                "tree_method": "find_mixed_case",
                "source_folder": "folder_with_multiple_files",
                "expected_value": ['folder_with_multiple_files/folder/file-EMPIAR-someting.tif']
            },
            {
                "tree_method": "find_accessions_in_names",
                "source_folder": "folder_with_multiple_files",
                "expected_value": ['folder_with_multiple_files/folder/file-EMPIAR-someting.tif']
            },
            {
                "tree_method": "find_with_date_names",
                "source_folder": "folder_with_date_name_files",
                "expected_value": [
                    'folder_with_date_name_files/prefix-12312000-suffix.txt',
                    'folder_with_date_name_files/prefix-2000:12:31-suffix.txt',
                    'folder_with_date_name_files/prefix-31:December:2000-suffix.txt',
                    'folder_with_date_name_files/prefix-31122000-suffix.txt',
                    'folder_with_date_name_files/prefix-31-Dec-2000-suffix.txt',
                    'folder_with_date_name_files/prefix-Dec-31-2000-suffix.txt',
                    'folder_with_date_name_files/prefix-2000-12-31-suffix.txt',
                    'folder_with_date_name_files/prefix-20001231-suffix.txt',
                    'folder_with_date_name_files/prefix-001231-suffix.txt',
                    'folder_with_date_name_files/prefix-31-December-2000-suffix.txt'
                ]
            },
            {
                "tree_method": "find_directories_with_mixed_files",
                "source_folder": "folder_with_multiple_file_types",
                "expected_value": ['folder_with_multiple_file_types/folder/']
            },
            {
                "tree_method": "find_long_names",
                "source_folder": "folder_with_long_name_folders",
                "expected_value": [
                    'folder_with_long_name_folders/a_folder_with_a_very_long_name_that_we_cannot_even_begin_to_comprehend/',
                    'folder_with_long_name_folders/folder/inner_folder/another_very_long_name_that_we_are_still_wondering_ever_found_the_light_of_day/',
                    'folder_with_long_name_folders/19jul19m_series0001_ite024_ang-51.0to51.0_thick2750_pxlsz1.331.dose_comp_a0.245_b-1.665_c2.81.bin4fourier_tomo3dSIRT_30_iters.mrc',
                ]
            },
            {
                "tree_method": "find_excessive_files_per_directory",
                "source_folder": "folder_with_multiple_folders",
                "expected_value": [f'folder_with_multiple_folders/folder7/']
            },
            {
                "tree_method": "find_obvious_directories",
                "source_folder": "single_empty_folder",
                "expected_value": ['single_empty_folder/folder/', 'single_empty_folder/folder/inner_folder/']
            },
            {
                "tree_method": "find_obvious_directories",
                "source_folder": "folder_with_non_ascii_characters",
                "expected_value": [
                    'folder_with_non_ascii_characters/dataset/',
                    'folder_with_non_ascii_characters/datasets/'
                ]
            },
            {
                "tree_method": "find_obvious_directories",
                "source_folder": "folder_with_multiple_file_types",
                "expected_value": [
                    'folder_with_multiple_file_types/folder/',
                    'folder_with_multiple_file_types/folder/files/',
                    'folder_with_multiple_file_types/folder/inner_folder/'
                ]
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "single_empty_folder",
                "expected_value": [
                    'single_empty_folder/',
                    'single_empty_folder/folder/',
                    'single_empty_folder/folder/inner_folder/'
                ]
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "single_empty_folder",
                "expected_value": [
                    'single_empty_folder/folder/',
                    'single_empty_folder/folder/inner_folder/'
                ],
                "kwargs": {"include_root": False}
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "empty_folder",
                "expected_value": ['empty_folder/', 'empty_folder/folder/'],
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "folder_with_multiple_folders",
                "expected_value": [],
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "folder_with_multiple_folders",
                "expected_value": [],
                "kwargs": {"include_root": False}
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "folder_with_multiple_file_types",
                "expected_value": [
                    'folder_with_multiple_file_types/',
                    'folder_with_multiple_file_types/folder/files/',
                    'folder_with_multiple_file_types/folder/inner_folder/'
                ],
            },
            {
                "tree_method": "find_empty_directories",
                "source_folder": "folder_with_multiple_file_types",
                "expected_value": [
                    'folder_with_multiple_file_types/folder/files/',
                    'folder_with_multiple_file_types/folder/inner_folder/'
                ],
                "kwargs": {"include_root": False}
            },
            {
                "tree_method": "find_odd_characters_in_names",
                "source_folder": "folder_with_long_name_folders",
                "expected_value": [
                    'folder_with_long_name_folders/a folder with & funny symbols in the ?? name/',
                    'folder_with_long_name_folders/a folder with spaces in the name/'
                ],
            },
            {
                "tree_method": "find_excessive_periods_in_names",
                "source_folder": "folder_with_long_name_folders",
                "expected_value": [
                    'folder_with_long_name_folders/a.folder.with.periods.in.the.name/',
                    'folder_with_long_name_folders/folder/a.file.with.many.periods.txt',
                    'folder_with_long_name_folders/19jul19m_series0001_ite024_ang-51.0to51.0_thick2750_pxlsz1.331.dose_comp_a0.245_b-1.665_c2.81.bin4fourier_tomo3dSIRT_30_iters.mrc'
                ],
            },
            {
                "tree_method": "find_external_references_in_names",
                "source_folder": "folder_with_long_name_folders",
                "expected_value": [
                    'folder_with_long_name_folders/folder/figure5.jpg',
                    'folder_with_long_name_folders/folder/supplementary-figure3a.jpg',
                ],
            },
            {
                "tree_method": "find_unknown_file_extensions",
                "source_folder": "folder_with_long_name_folders",
                "expected_value": [
                    'folder_with_long_name_folders/folder/file.wrx',
                    'folder_with_long_name_folders/folder/file.onx',
                    'folder_with_long_name_folders/folder/file.dog',
                ],
            },
            {
                "tree_method": "find_non_ascii_characters",
                "source_folder": "folder_with_non_ascii_characters",
                "expected_value": [
                    'wïth_ñõn_æšçiį'
                ]
            }
        ]
        for data_dict in data:
            dir_entries = utils.scandir_recursive(TEST_DATA / data_dict["source_folder"])
            tree = models.Tree.from_data(dir_entries, prefix=str(TEST_DATA), args=args)
            if "kwargs" in data_dict:
                result = sorted(
                    getattr(
                        tree,
                        data_dict["tree_method"]
                    )(**data_dict["kwargs"])
                )
            else:
                result = sorted(getattr(tree, data_dict["tree_method"])())
            print(data_dict["tree_method"])
            # print(result)
            # print(data_dict['expected_value'])
            self.assertEqual(sorted(data_dict["expected_value"]), result)


class TestView(Tests):
    def test_view_tree(self):
        """Test that we can display the tree"""
        args = cli.cli(f"bandbox view {TEST_DATA / 'empty_folder'}")
        self.assertEqual(str(BASE_DIR / "bandbox.cfg"), args.config_file)
        self.assertIsInstance(args._configs, configparser.ConfigParser)
        sys.stdout = io.StringIO()
        managers.view(args)
        self.assertRegex(sys.stdout.getvalue(), r"(?s).*empty_folder.*folder.*")


class TestAnalyse(Tests):
    def test_analyse_all_engines(self):
        """Run all engines"""
        args = cli.cli(f"bandbox analyse {TEST_DATA / 'empty_folder'}")
        sys.stdout = io.StringIO()
        managers.analyse(args)
        self.assertRegex(sys.stdout.getvalue(),
                         r"(?s).*(unknown file extensions|accessions in names|entities with dates).*")


class TestUtils(Tests):
    def test_scandir_recursive(self):
        """Test that we can recursively scan a directory"""
        base_dir = TEST_DATA / "folder_with_single_file"
        path_generator = utils.scandir_recursive(base_dir)
        self.assertIsInstance(path_generator, types.GeneratorType)
        for dir_entry in path_generator:
            print(dir_entry, dir_entry.path)



import os
import sys

import bandbox
from bandbox import managers
from bandbox.cli import parse_args


def main():
    args = parse_args()
    if args is None:
        return os.EX_DATAERR
    try:
        if args.command == 'analyse':
            return managers.analyse(args)
        elif args.command == 'view':
            return managers.view(args)
    except KeyboardInterrupt:
        pass
    return os.EX_OK


if __name__ == '__main__':
    sys.exit(main())



import argparse
import configparser
import os
import pathlib
import shlex
import sys
from typing import Union, Iterable, Optional, List

# options
hide_file_counts = {
    'args': ['--hide-file-counts'],
    'kwargs': dict(
        default=True,
        action='store_false',
        help="display file counts [default: True]"
    )
}
path = {
    'args': ['path'],
    'kwargs': {
        'nargs': '?',
        'default': '',
        'type': pathlib.Path,
        'help': "path to diagnose [default: '.']"
    }
}

prefix = {
    'args': ['-p', '--prefix'],
    'kwargs': {
        'default': '',
        'help': "prefix to exclude [default: '']"
    }
}


def _add_arg(parser_: argparse.ArgumentParser, option: dict):
    """Add options to parser"""
    return parser_.add_argument(*option['args'], **option['kwargs'])


class LocalConfigParser(configparser.ConfigParser):
    """String-printable version"""

    def __init__(self, *args, **kwargs):
        super().__init__(
            interpolation=configparser.ExtendedInterpolation(),
            converters={
                'list': self.getlist,
                'tuple': self.gettuple,
                'python': self.getpython,
                'cre': self.getcre,
            }, *args, **kwargs)
        self._filenames = None

    @property
    def filenames(self):
        return self._filenames

    def read(self, filenames: Union[os.PathLike, Iterable[os.PathLike]], encoding: Optional[str] = None) -> None:
        self._filenames = filenames
        super().read(filenames, encoding)

    def __str__(self):
        string = ""
        for section in self.sections():
            string += f"[{section}]\n"
            for option in self[section]:
                string += f"{option} = {self.get(section, option, raw=True)}\n"
            string += "\n"
        return string

    @staticmethod
    def getlist(value):
        """Convert the option value to a list"""
        return list(map(lambda s: s.strip(), value.split(',')))

    @staticmethod
    def gettuple(value):
        """Convert the option value to a tuple"""
        return tuple(map(lambda s: s.strip(), value.split(',')))

    @staticmethod
    def getpython(value):
        """Evaluate the option value as literal Python code"""
        return eval(value)

    @staticmethod
    def getcre(value):
        """Evaluate a compiled regular expression"""
        import re
        return re.compile(value)


parser = argparse.ArgumentParser(prog='bandbox', description="Evaluate how organised your dataset is")

parent_parser = argparse.ArgumentParser(add_help=False)
parent_parser.add_argument('-c', '--config-file',
                           help="oil configs [default: value of BANDBOX_CONFIG environment variable")
parent_parser.add_argument('-v', '--verbose', default=False, action='store_true',
                           help="verbose output to terminal in addition to log files [default: False]")

subparsers = parser.add_subparsers(dest='command', title='Commands available')

# analyse
analyse_parser = subparsers.add_parser(
    'analyse',
    description='analyse the data tree',
    help='analyse the dataset',
    parents=[parent_parser]
)
_add_arg(analyse_parser, path)
analyse_parser.add_argument('--include-root', default=False, action='store_true',
                            help="include the root directory for analysis [default: False]")
_add_arg(analyse_parser, prefix)
analyse_parser.add_argument('-T', '--show-tree', default=False, action='store_true',
                            help="display the tree [default: False]")
SUMMARY_SIZE = 5
analyse_parser.add_argument(
    '-s', '--summarise-size',
    default=SUMMARY_SIZE, type=int,
    help=f"summarise size [default: {SUMMARY_SIZE}]"
)
analyse_parser.add_argument(
    '-S', '--summarise',
    default=False,
    action='store_true',
    help=f"summarise to --summary-size ({SUMMARY_SIZE}) results in each group [default: False]"
)
_add_arg(analyse_parser, hide_file_counts)

# view
view_parser = subparsers.add_parser(
    'view',
    description='display the data tree',
    help='display the data tree',
    parents=[parent_parser]
)
_add_arg(view_parser, path)
_add_arg(view_parser, prefix)
# todo: remove
# view_parser.add_argument('-v', '--verbose', default=False, action='store_true',
#                          help="verbose output which will display all the directories found [default: False]")
view_parser.add_argument('-f', '--input-file', help="input data from a file")
_add_arg(view_parser, hide_file_counts)


def parse_args():
    """Parse CLI args"""
    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
        return None
    # config-file
    if args.config_file is None:
        try:
            args.config_file = os.environ['BANDBOX_CONFIG']
        except KeyError:
            print(f"error: no configs found; please set BANDBOX_CONFIG envvar or provide --config-file path",
                  file=sys.stderr)
            print(f"info: copy and modify the config file from https://raw.githubusercontent.com/emdb-empiar/bandbox/master/bandbox.cfg", file=sys.stderr)
            return None
    # read configs
    configs = LocalConfigParser()
    configs.read(args.config_file)
    args._configs = configs
    # the path must exist
    if not args.path.exists():
        print(f"error: invalid path '{args.path}'", file=sys.stderr)
        return None
    return args


def cli(cmd):
    """CLI standin"""
    sys.argv = shlex.split(cmd)
    return parse_args()



import os
import re
import sys
from collections import UserDict

import bandbox


class Tree(UserDict):
    def __new__(cls, sep='/', show_file_counts=True):
        cls.sep = sep
        cls.show_file_counts = show_file_counts
        return super().__new__(cls)

    @classmethod
    def from_data(cls, data, prefix="", show_file_counts=True, args=None):
        tree = cls()
        tree._args = args
        tree._configs = args._configs
        tree.show_file_counts = show_file_counts
        for t in data:
            tree.insert(t, prefix=prefix)
        return tree

    def insert(self, dir_entry: os.DirEntry, prefix: str = ''):
        """Insert the path into the tree creating nodes if necessary"""
        if prefix == '.':
            path_list = dir_entry.path.strip(self.sep).split(self.sep)
        else:
            path_list = dir_entry.path[len(prefix):].strip(self.sep).split(self.sep)
        # first, deal with directories
        insertion_point = self.data
        for element in path_list[:-1]:
            if element not in insertion_point:
                insertion_point[element] = dict()
                insertion_point = insertion_point[element]
            else:
                insertion_point = insertion_point[element]
        # last item
        if dir_entry.is_file():
            if '_files' not in insertion_point:
                insertion_point['_files'] = [path_list[-1]]
            else:
                insertion_point['_files'] += [path_list[-1]]
        else:
            insertion_point[path_list[-1]] = dict()

    # @staticmethod
    def file_counts(self, file_list):
        file_counts = dict()
        for file_ in file_list:
            file_match = self._configs.getcre('regex', 'file_re').match(file_)
            if file_match:
                ext = file_.split('.')[-1]
                if ext not in file_counts:
                    file_counts[ext] = 1
                else:
                    file_counts[ext] += 1
        return file_counts

    def _recursive_string(self, extraction_point, indent=""):
        string = ""
        for key, value in extraction_point.items():
            if isinstance(value, dict):
                string += f"{indent}└── {key}\n"
                indent += "\t"
                string += f" {self._recursive_string(extraction_point=extraction_point[key], indent=indent)}"
                indent = indent[:-1]
            else:
                if len(value) == 1:
                    item = "file"
                else:
                    item = "files"
                if self.show_file_counts:
                    file_counts_str = ""
                    file_counts = self.file_counts(value)
                    for ext, count in file_counts.items():
                        file_counts_str += f"{ext}={count}; "
                    string += f"{indent}└── [{len(value)} {item}: {file_counts_str}]\n"
                else:
                    string += f"{indent}└── [{len(value)} {item}]\n"

        return string

    def __str__(self):
        """Print as tree"""
        string = self._recursive_string(self.data)
        return string

    @staticmethod
    def evaluate_predicate(tree_dict, predicate, parent=""):
        """Generic rule to evaluate elements of the tree"""
        output = list()
        for dir_entry, children_dict in tree_dict.items():
            # each method that finds something has to provide an assertion method which takes
            # - dir_entry: the current directory entry
            # - children_dict: any children dictionary associated with the current directory
            # - tree_dict: the parent directory dictionary
            # - parent_path: string with the path to the dir_entry:
            output += predicate(dir_entry, children_dict, tree_dict, parent)
            if isinstance(children_dict, (dict, Tree)):
                output += Tree.evaluate_predicate(children_dict, predicate, parent=f"{parent}{dir_entry}/")
        return output

    def find_empty_directories(self, include_root=True) -> list:
        """Identify directories with no files"""

        def empty_directories_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if len(children_dict) == 0:  # terminal empty folder
                output.append(f"{parent_path}{dir_entry}/")
            if len(children_dict) == 1:  # non-terminal folder with files only
                if "_files" not in children_dict and not isinstance(children_dict, list):
                    output.append(f"{parent_path}{dir_entry}/")
            return output

        empty_dirs = Tree.evaluate_predicate(self, empty_directories_predicate)
        if include_root:
            return empty_dirs
        return empty_dirs[1:]

    def find_obvious_directories(self, include_root=True) -> list:
        """Identify directories with obvious names"""

        def obvious_directories_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if self._configs.getcre('regex', 'obvious_files_re').match(dir_entry):
                output.append(f"{parent_path}{dir_entry}/")
            return output

        obvious_directories = Tree.evaluate_predicate(self, obvious_directories_predicate)
        return obvious_directories

    def find_excessive_files_per_directory(self) -> list:
        """Identify directories with excessive files"""

        def excessive_files_per_directory_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':  # in parent_dict:
                if len(parent_dict['_files']) > self._configs.getint('bandbox', 'max_files'):
                    # fixme: presentation logic bled in ---> remove
                    # dir_name = parent_path.ljust(LEFT_COL_WIDTH)
                    # num_files = f"[{len(parent_dict['_files'])} files]"
                    # output.append(f"{dir_name}{num_files.rjust(RIGHT_COL_WIDTH - 3)}")
                    output.append(f"{parent_path}")
            return output

        excess = Tree.evaluate_predicate(self, excessive_files_per_directory_predicate)
        return excess

    def find_long_names(self) -> list:
        """Identify path elements with long names"""

        def long_names_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if len(file) > self._configs.getint('bandbox', 'max_name_length'):
                        output.append(f"{parent_path}{file}")
            if len(dir_entry) > self._configs.getint('bandbox', 'max_name_length'):
                output.append(f"{parent_path}{dir_entry}/")
            return output

        long_names = Tree.evaluate_predicate(self, long_names_predicate)
        return long_names

    def find_directories_with_mixed_files(self) -> list:
        def directories_with_mixed_files_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                files = self.file_counts(parent_dict['_files'])
                if len(files) > 1:
                    output.append(f"{parent_path}")
            return output

        mixed_dirs = Tree.evaluate_predicate(self, directories_with_mixed_files_predicate)
        return mixed_dirs

    def find_with_date_names(self) -> list:

        def date_names_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    for date_re in map(lambda r: re.compile(r), self._configs.getlist('regex', 'date_re')):
                        if date_re.match(file):
                            output.append(f"{parent_path}{file}")
            return output

        date_names = list(set(Tree.evaluate_predicate(self, date_names_predicate)))
        return date_names

    def find_accessions_in_names(self) -> list:
        def accessions_in_names_predicate(dir_entry, chidren_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if self._configs.getcre('regex', 'accession_names_re').match(file):
                        output.append(f"{parent_path}{file}")
            return output

        accession_names = Tree.evaluate_predicate(self, accessions_in_names_predicate)
        return accession_names

    def find_mixed_case(self) -> list:
        """Find elements of the tree with names having mixed case"""

        def mixed_case_predicate(dir_entry, children_dict, parent_dict,
                                 parent_path):  # assertion callback for this 'find...' method
            output = list()
            if dir_entry == '_files':  # file contents
                for file in parent_dict['_files']:
                    if re.match(r".*[A-Z].*", file) and re.match(r".*[a-z].*", file):  # mixed case
                        output.append(f"{parent_path}{file}")
            if re.match(r".*[A-Z].*", dir_entry) and re.match(r".*[a-z].*", dir_entry):  # if the dirname has mixed case
                output.append(f"{parent_path}{dir_entry}/")
            return output

        mixed_case = Tree.evaluate_predicate(self, mixed_case_predicate)
        return mixed_case

    def find_odd_characters_in_names(self):
        """Find odd characters in path components"""

        def odd_characters_in_names_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if self._configs.getcre('regex', 'odd_chars_re').match(file):
                        output.append(f"{parent_path}{file}")
            if self._configs.getcre('regex', 'odd_chars_re').match(dir_entry):
                output.append(f"{parent_path}{dir_entry}/")
            return output

        odd_characters = Tree.evaluate_predicate(self, odd_characters_in_names_predicate)
        return odd_characters

    def find_excessive_periods_in_names(self):
        """Find odd characters in path components"""

        def excessive_periods_in_names_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if self._configs.getcre('regex', 'periods_in_name_fewer_than_re').match(file):
                        output.append(f"{parent_path}{file}")
            if self._configs.getcre('regex', 'periods_in_name_fewer_than_re').match(dir_entry):
                output.append(f"{parent_path}{dir_entry}/")
            return output

        odd_characters = Tree.evaluate_predicate(self, excessive_periods_in_names_predicate)
        return odd_characters

    def find_external_references_in_names(self):
        """Find external references in names"""

        def external_references_in_names_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if self._configs.getcre('regex', 'external_refs_re').match(file):
                        output.append(f"{parent_path}{file}")
            if self._configs.getcre('regex', 'external_refs_re').match(dir_entry):
                output.append(f"{parent_path}{dir_entry}/")
            return output

        external_refs = Tree.evaluate_predicate(self, external_references_in_names_predicate)
        return external_refs

    def find_unknown_file_extensions(self):
        """Find unknown file extensions"""
        def unknown_file_extensions_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if not self._configs.getcre('regex', 'file_extension_re').match(file):
                        output.append(f"{parent_path}{file}")
            return output

        unknown_exts = Tree.evaluate_predicate(self, unknown_file_extensions_predicate)
        return unknown_exts

    def find_non_ascii_characters(self):
        """Find files/folders with non-ascii characters

        The heuristic is that if the length of the string is equal to utf-8-encoded byte string then it's ascii
        """
        def non_ascii_character_predicate(dir_entry, children_dict, parent_dict, parent_path):
            output = list()
            if dir_entry == '_files':
                for file in parent_dict['_files']:
                    if len(file) != len(file.encode('utf-8')):
                        output.append(file)
            if len(dir_entry) != len(dir_entry.encode('utf-8')):
                output.append(dir_entry)
            return output

        non_ascii_characters = Tree.evaluate_predicate(self, non_ascii_character_predicate)
        return non_ascii_characters



import asyncio
import json
import sys

from bandbox import utils
from bandbox.models import Tree


async def _analyse_engines(tree, args):
    """Run engines concurrently"""
    from bandbox import engines
    import inspect
    engines_ = inspect.getmembers(engines, inspect.isfunction)
    awaitables = list()
    for engine_name, engine in engines_:
        if engine_name.startswith('_'):  # skip private methods
            continue
        awaitables.append(engine(tree, args))
    await asyncio.gather(*awaitables)


def analyse(args):
    """Analyse the given dataset"""
    # decide which engines we will include
    # e.g. n2_long_names -> list of entities with long names
    # entry point
    dir_entries = utils.scandir_recursive(args.path)
    tree = Tree.from_data(dir_entries, prefix=str(args.prefix), show_file_counts=args.hide_file_counts, args=args)
    if args.show_tree:
        print(tree)
    if sys.version_info.minor > 6:  # 3.7+
        asyncio.run(_analyse_engines(tree, args))
    else:  # python 3.6
        loop = asyncio.get_event_loop()
        loop.run_until_complete(_analyse_engines(tree, args))
        loop.close()


def view(args):
    """View the given dataset"""
    if args.input_file:
        with open(args.input_file) as f:
            data = f.read().strip().split(', ')
    else:
        data = utils.scandir_recursive(args.path)
    tree = Tree.from_data(data, prefix=str(args.path.parent), show_file_counts=args.hide_file_counts, args=args)
    if args.verbose:
        print(f"info: displaying nested tree data...", file=sys.stderr)
        print(json.dumps(tree.data, indent=4), file=sys.stderr)
    try:
        print(tree)
    except BrokenPipeError:
        pass
